@@ -1,4 +1,5 @@
 use crate::*;
+use std::collections::HashMap;
 
 
 static DEBUG_CANDIDATE_RESOLUTION: bool = false;
@@ -18,6 +19,7 @@ pub struct State
 	pub bankdata: Vec<asm::BankData>,
 	pub symbols: asm::SymbolManager,
 	pub symbol_guesses: asm::SymbolManager,
+	pub instruction_size_guesses: HashMap<diagn::Span, usize>,
 	pub rulesets: Vec<asm::Ruleset>,
 	pub active_rulesets: Vec<RulesetRef>,
 	pub cur_bank: BankRef,
@@ -95,6 +97,7 @@ impl Assembler
         -> Result<AssemblyOutput, ()>
 	{
 		let mut symbol_guesses = asm::SymbolManager::new();
+		let mut instruction_size_guesses = HashMap::<diagn::Span, usize>::new();
 
 		let mut iteration = 0;
 		loop
@@ -102,12 +105,14 @@ impl Assembler
 			self.state = State::new();
 			self.state.is_first_pass = iteration == 0;
 			std::mem::swap(&mut self.state.symbol_guesses, &mut symbol_guesses);
+			std::mem::swap(&mut self.state.instruction_size_guesses, &mut instruction_size_guesses);
 
 			iteration += 1;
 			//dbg!(iteration);
 
 			//dbg!(&symbol_guesses);
 			//dbg!(&self.state.symbols);
+			//dbg!(&self.state.instruction_size_guesses);
 
 			let pass_report = diagn::RcReport::new();
 
@@ -143,7 +148,8 @@ impl Assembler
 					pass_report.clone(),
 					bank,
 					bankdata,
-					fileserver);
+					fileserver,
+					&mut instruction_size_guesses);
 					
 				if pass_report.has_errors() || !bank_output.is_ok()
 				{
@@ -202,6 +208,7 @@ impl State
 			bankdata: Vec::new(),
 			symbols: asm::SymbolManager::new(),
 			symbol_guesses: asm::SymbolManager::new(),
+			instruction_size_guesses: HashMap::new(),
 			rulesets: Vec::new(),
 			active_rulesets: Vec::new(),
 			cur_bank: BankRef { index: 0 },
@@ -430,7 +437,8 @@ impl State
 		report: diagn::RcReport,
 		bank: &asm::Bank,
 		bankdata: &asm::BankData,
-		fileserver: &dyn util::FileServer)
+		fileserver: &dyn util::FileServer,
+		instruction_size_guesses: &mut HashMap<diagn::Span, usize>)
 		-> Result<util::BitVec, ()>
 	{
 		let mut bitvec = util::BitVec::new();
@@ -500,14 +508,22 @@ impl State
 				_ => unreachable!(),
 			};
 
-			let (bigint, size) = match resolved
+			let (bigint, size) = match resolved.get_bigint()
 			{
-				expr::Value::Integer(bigint) =>
+				Some(bigint) =>
 				{
 					match bigint.size
 					{
 						Some(size) =>
 						{
+							// Set latest resolved value as the size guess
+							// for next iterations, but only if the early size guess
+							// couldn't be inferred.
+							if invoc.size_guess == 0
+							{
+								instruction_size_guesses.insert(invoc.span.clone(), size);
+							}
+							
 							if size == invoc.size_guess
 							{
 								(bigint, size)
@@ -536,7 +552,7 @@ impl State
 					}
 				}
 
-				_ =>
+				None =>
 				{
 					report.error_span(
 						format!(
@@ -1036,7 +1052,7 @@ impl State
 			}
 			else if self.is_first_pass
 			{
-				Ok(expr::Value::make_integer(0))
+				Ok(expr::Value::Unknown)
 			}
 			else
 			{
@@ -1077,52 +1093,88 @@ impl State
 			expr::Value::Bool(value) => Ok(value),
 			_ =>
 			{
-				info.report.error_span("expected boolean argument", info.span);
+				info.report.error_span("expected boolean argument", &info.arg_spans[index]);
 				Err(true)
 			}
 		}
 	}
 
 
-	fn eval_fn_get_bigint_arg<'a>(
+	fn eval_fn_check_unknown_arg<'a>(
+		info: &'a expr::EvalFunctionInfo,
+		index: usize,
+		is_first_pass: bool)
+		-> bool
+	{
+		if !is_first_pass
+		{
+			return false;
+		}
+		
+		match &info.args[index]
+		{
+			expr::Value::Unknown => true,
+			_ => false,
+		}
+	}
+
+
+	fn eval_fn_get_string_arg<'a>(
 		info: &'a expr::EvalFunctionInfo,
 		index: usize)
-		-> Result<&'a util::BigInt, bool>
+		-> Result<&'a expr::ValueString, bool>
 	{
 		match &info.args[index]
 		{
-			expr::Value::Integer(value) => Ok(value),
+			expr::Value::String(value) => Ok(value),
 			_ =>
 			{
-				info.report.error_span("expected integer argument", info.span);
+				info.report.error_span("expected string argument", &info.arg_spans[index]);
 				Err(true)
 			}
 		}
 	}
 
 
-	fn eval_fn_get_sized_bigint_arg<'a>(
-		info: &'a expr::EvalFunctionInfo,
+	fn eval_fn_get_bigint_arg(
+		info: &expr::EvalFunctionInfo,
 		index: usize)
-		-> Result<&'a util::BigInt, bool>
+		-> Result<util::BigInt, bool>
 	{
-		match &info.args[index]
+		match info.args[index].get_bigint()
 		{
-			expr::Value::Integer(value) =>
+			Some(value) => Ok(value),
+			None =>
+			{
+				info.report.error_span("expected integer argument", &info.arg_spans[index]);
+				Err(true)
+			}
+		}
+	}
+
+
+	fn eval_fn_get_sized_bigint_arg(
+		info: &expr::EvalFunctionInfo,
+		index: usize)
+		-> Result<util::BigInt, bool>
+	{
+		match info.args[index].get_bigint()
+		{
+			Some(value) =>
 			{
 				match value.size
 				{
 					Some(_) => Ok(value),
 					None =>
 					{
-						info.report.error_span("unsized integer argument", info.span);
+						info.report.error_span("unsized integer argument", &info.arg_spans[index]);
 						Err(true)
 					}
 				}
 			}
-			_ =>
+			None =>
 			{
-				info.report.error_span("expected integer argument", info.span);
+				info.report.error_span("expected integer argument", &info.arg_spans[index]);
 				Err(true)
 			}
 		}
@@ -1178,12 +1230,16 @@ impl State
 					"inchexstr" =>
 					{
 						State::eval_fn_check_arg_number(info, 1)?;
-						let bigint = State::eval_fn_get_bigint_arg(info, 0)?;
-						let filename = bigint.as_string();
+						if State::eval_fn_check_unknown_arg(info, 0, self.is_first_pass)
+						{
+							return Ok(expr::Value::make_integer(util::BigInt::new_from_str("")));
+						}
+
+						let value_string = State::eval_fn_get_string_arg(info, 0)?;
 						let new_filename = util::filename_navigate(
 							info.report.clone(),
 							&ctx.cur_filename,
-							&filename,
+							&value_string.utf8_contents,
 							&info.span)
 							.map_err(|_| true)?;
 
@@ -1336,10 +1392,12 @@ impl State
 				fileserver,
 				true,
 				info.args)?;
+			
+			//println!("  value = {:?}", value);
 				
-			let (bigint, size) = match value
+			let (bigint, size) = match value.get_bigint()
 			{
-				expr::Value::Integer(bigint) =>
+				Some(bigint) =>
 				{
 					match bigint.size
 					{
@@ -1383,6 +1441,7 @@ impl State
 			parser.expect_linebreak()?;
 		}
 
+		//println!("  result size = {:?}", result.size);
 		Ok(expr::Value::make_integer(result))
 	}
 }
